@@ -1,35 +1,49 @@
-/* (C) Matthew Kirkwood <matthew@hairy.beasts.org> */
+/* (C) Matthew Kirkwood <matthew@hairy.beasts.org>
+   (C) 2002 Rusty Russell IBM <rusty@rustcorp.com.au>
+ */
 
-/* Atomic dec: return 1 unless counter decremented to zero. */
-/* Taken from 2.4.18's linux/include/asm-i386/atomic.h */
-static int __futex_down(int *counter)
+/* Atomic dec: return new value. */
+static __inline__ int __futex_down(int *counter)
 {
-	unsigned char c;
+	int val;
+	unsigned char eqz;
 
+	/* Don't decrement if already negative. */
+	val = *counter;
+	if (val < 0)
+		return val;
+
+	/* Damn 386: no cmpxchg... */
 	__asm__ __volatile__(
 		"lock; decl %0; sete %1"
-		:"=m" (*counter), "=qm" (c)
+		:"=m" (*counter), "=qm" (eqz)
 		:"m" (*counter) : "memory");
-	return c == 0;
+
+	/* We know if it's zero... */
+	if (eqz) return 0;
+	/* Otherwise, we have no way of knowing value.  Guess -1 (if
+           we're wrong we'll spin). */
+	return -1;
 }
 
-/* Atomic inc: return 1 unless counter incremented from 0 to 1. */
+/* Atomic inc: return 1 if counter incremented from 0 to 1. */
 static __inline__ int __futex_up(int *c)
 {
-	int r = 0;
+	int r = 1;
 
-	/* This actually tests if result >= 1. --RR */
+	/* This actually tests if result >= 1.  Damn 386. --RR */
 	__asm__ __volatile__ (
 	"	lock; incl %1\n"
 	"	jg 1f\n"
-	"	incl %0\n"
+	"	decl %0\n"
 	"1:\n"
 		: "=q"(r), "=m"(*c) : "0"(r)
 	);
 	return r;
 }
 
-static __inline__ void __furwock_inc(int *c)
+/* Simple atomic increment. */
+static __inline__ void __atomic_inc(int *c)
 {
 	__asm__ __volatile__(
 		"lock; incl %0"
@@ -37,6 +51,7 @@ static __inline__ void __furwock_inc(int *c)
 		:"m" (*c));
 }
 
+/* Atomic decrement, and return 1 if result is negative. */
 static __inline__ int __furwock_dec_negative(int *c)
 {
 	unsigned char r;
@@ -46,7 +61,7 @@ static __inline__ int __furwock_dec_negative(int *c)
 		:"=m" (*c), "=qm" (r)
 		:"m" (*c) : "memory");
 
-	return r != 0;
+	return r;
 }
 
 /* Commit the write, so it happens before we send the semaphore to
